#!/usr/bin/env python

"""
kaula.py
Computation of Kaula expansion.

References:

Walterova (2021), PhD Thesis, Charles Univ.
Ellis & Murray (2000), Icarus, 147, 129-144.
Iszak etal. (1964), SAO Spec. Rep., 140.
Renaud etal. (2021), Planet. Sci. J., 2, 4.

"""

__author__ = "Miroslav Broz (miroslav.broz@email.cz)"
__version__ = "May 19th 2021"

import sys
from math import sin, cos
import poly
import poly2

def fact(k):
  """Factorial"""

  f = 1
  for i in xrange(1, k+1):
    f *= i
  return f

def C_f(n, k):
  """Combination number, incl. non-integer values"""

#  if k > n:
#    return 0

  C = 1
  for i in xrange(0, k):
    C *= (n-i)

  C /= fact(k)
  return C

def X_f(a, b, c, d):
  """Newcomb operators X^ab_cd; Ellis & Murray (2000)."""

#  print "call X_f(", a, b, c, d, ")"
  # Eq. (23)
  if c == 0 and d == 0:
    X = 1

  # Eq. (24)
  elif c == 1 and d == 0:
    X = b - a/2.

  elif c < 0 or d < 0:
    X = 0

  elif d > c:
    X = X_f(a, -b, d, c)

  # Eq. (25)
  elif d == 0:
    X = 2*(2*b - a)*X_f(a, b+1, c-1, 0) + (b - a)*X_f(a, b+2, c-2, 0)
    X /= 4.*c

  # Eq. (26)
  else:
    s = 0.
    maxj = max(c, d)
    for j in xrange(2, maxj+1):
      s += (-1)**j * C_f(3./2., j) * X_f(a, b, c-j, d-j)

    X = -2*(2*b + a)*X_f(a, b-1, c, d-1) - (b + a)*X_f(a, b-2, c, d-2)
    X -= (c - 5*d + 4 + 4*b + a)*X_f(a, b, c-1, d-1)
    X += 2*(c - d + b)*s
    X /= 4.*d

  return X

def Ha_f(a, b, c, e = 0.1, maxsigma = 3, debug = True, fort = False):
  """Hansen coefficients X^ab_c; Ellis & Murray (2000), Eq. (22)."""
 
#  print "call H(", a, b, c, ")"
  alpha = max(0, c - b)
  beta = max(0, b - c)
  H = 0.

  if debug:
    p = poly.Poly(n = 2*maxsigma+abs(c-b)+1, variable = "e")

  for sigma in xrange(0, maxsigma+1):
    X = X_f(a, b, sigma+alpha, sigma+beta)
    fac = 2*sigma + abs(c - b)
    H += X * e**fac

    if debug:
      #print "X(%2d, %2d, %2d, %2d) = %12.8f  for e^%2d = %.4e  ->  Ha = %.8f" % (a, b, sigma+alpha, sigma+beta, X, fac, e**fac, H)
      p.a[fac] = X

  if debug:
#    print "Ha(", a, ", ", b, ", ", c, ") = ", H
    print ""
#    print "polynomial  : ", p
    p2 = poly.mul(p, p)
    print "polynomial^2: ", p2
    print "evaluate    : ", p2.evaluate(e)
    print ""

  if fort:
    f = open("kaula5.fort", "a")
    f.write(p2.fortran() + '\n')
    f.close()

  return H

def G_f(l, p, q, e = 0.1, m = 2, maxsigma = 3, debug = True, fort = False):
  """Kaula eccentricity functions G_lpq; Walterova (2021), Eq. (1.44)."""

  if debug:
    print "--"
    print "l = ", l
    print "m = ", m
    print "p = ", p
    print "q = ", q
    print "omega_", l, m, p, q, " = ", l-2*p+q, " n - ", m, " omega"

  G = Ha_f(-l-1, l-2*p, l-2*p+q, e = e, maxsigma = maxsigma, debug = debug, fort = fort)

#  if debug:
#    print "G(", l, ", ", p, ", ", q, ") = ", G
#    print "G^2 = ", G*G
  return G

def H_f(l, p, q, e = 0.1, m = 2, maxsigma = 3, debug = True):
  """Kaula eccentricity functions H_lpq; Walterova (2021), Eq. (1.45)."""

  if debug:
    print "--"
    print "l = ", l
    print "m = ", m
    print "p = ", p
    print "q = ", q
    print "omega_", l, m, p, q, " = ", l-2*p+q, " n - ", m, " omega"

  H = Ha_f(l, l-2*p, l-2*p+q, e = e, maxsigma = maxsigma, debug = debug)

#  if debug:
#    print "H(", l, ", ", p, ", ", q, ") = ", H
#    print "H^2 = ", H*H
  return H

def F_f(l, m, p, i = 0.1, q = 0, debug = True):
  """Kaula inclination functions F_lmp"""
  print "--"
  print "l = ", l
  print "m = ", m
  print "p = ", p
  print "q = ", q
  print "omega_", l, m, p, q, " = ", l-2*p+q, " n - ", m, " omega"

  delta = int((l-m+1)/2)
  j1 = max(0, l-m-2*p)
  j2 = min(l-m, 2*l-2*p)
  c = cos(i/2.0)
  s = sin(i/2.0)
  fac = (-1)**delta * fact(l+m)/(2.**l*fact(p)*fact(l-p))
  F = 0.

  if debug:
    P = poly2.Poly2(n = 0, variable = ("s", "c"))

  for j in xrange(j1,j2+1):
    alpha = m-l+2*p+2*j
    beta = 2*l-alpha
    fac2 = fac * (-1)**j * C_f(2*l-2*p, j) * C_f(2*p, l-m-j)
    F += fac2 * s**alpha * c**beta

    if debug:
      P.a.append(fac2)
      P.e.append((alpha, beta))

  if debug:
#    print "F(", l, ", ", m, ", ", p, ") = ", F
    print ""
#    print "polynomial   = ", P
    P2 = poly2.mul(P, P)
    print "polynomial^2 = ", P2
    print "evaluate     = ", P2.evaluate(s, c)
    print ""

  return F

def test():
  """Test elementary functions"""

  print "10! = ", fact(10)
  print "6! = ", fact(6)
  print "4! = ", fact(4)
  print "10!/(4!6!) = ", fact(10)/(fact(4)*fact(10-4))
  print "C(10, 4) = ", C_f(10, 4)
  print "C(10, 6) = ", C_f(10, 6)
  print "--"

  print "cf. Izsak (1964), p. 55"
  print "X(1,1,0,0) = ", X_f(1,1,0,0); print "cf. 1"
  print "X(1,0,1,0) = ", X_f(1,0,1,0); n=1; k=0; print "cf. ", k-0.5*n
  print "X(2,1,2,0) = ", X_f(2,1,2,0); n=2; k=1; print "cf. ", 0.625*k + 0.5*k**2 + (-0.375 - 0.5*k)*n + 0.125*n**2
  print "X(2,2,1,1) = ", X_f(2,2,1,1); n=2; k=2; print "cf. ", -k**2 + 0.25*n + 0.25*n**2
  print "X(3,2,3,0) = ", X_f(3,2,3,0); n=3; k=2; print "cf. ", 0.54166666*k + 0.625*k**2 + 0.16666666*k**3 + (-0.3541666 - 0.6875*k - 0.25*k**2)*n + (0.1875 + 0.125*k)*n**2 - 0.20833333e-1*n**3
  print "--"

  e = 0.1
  print "e = ", e
  print "cf. Ellis & Murray (2000)"
  Ha_f( 0,  3,  3, e = e); print "cf. 1 - 9 e^2                                          (Eq. 73)"
  Ha_f(-1,  3,  4, e = e); print "cf. 7/2 e - 179/8 e^3                                  (Eq. 74)"
  Ha_f( 1,  3,  3, e = e); print "cf. 1 - 17/2 e^2                                       (Eq. 75)"
  Ha_f(-2,  3,  4, e = e); print "cf. 4 e - 24 e^3                                       (Eq. 76)"
  Ha_f( 2,  3,  3, e = e); print "cf. 1 - 15/2 e^2                                       (Eq. 77)"
  Ha_f(-4,  3,  4, e = e); print "cf. 5 e - 22 e^3                                       (Eq. 80)"

  print "cf. Walterova (2021)"
  G_f(2, 0,  0, e = e); print "cf. 1 - 5 e^2 = ", 1.-5.*e**2, "  (Eq. 3.4)"
  G_f(2, 0, -1, e = e); print "cf. 0 + 1/4 e^2 = ", 1./4.*e**2
  G_f(2, 0, +1, e = e); print "cf. 0 + 49/4 e^2 = ", 49./4.*e**2
  G_f(2, 0, -2, e = e);
  G_f(2, 0, +2, e = e);

def test_renaud1():
  """Test orders up to e^10 wrt. Renaud etal. (2021)."""
  e = 0.4
  print "e = ", e
  print "cf. Renaud etal. (2021), Tab. 2, p. 11"
  G_f(2, 0, -5, e = e); print "cf. 0 + 6561/1638400 e^10"
  G_f(2, 0, -4, e = e); print "cf. 0 + 1/576 e^8 + 7/2880 e^10"
  G_f(2, 0, -3, e = e); print "cf. 0 + 1/2304 e^6 + 11/18432 e^8 + 619/983040 e^10"
  G_f(2, 0, -2, e = e); print "cf. 0"
  G_f(2, 0, -1, e = e); print "cf. 0 + 1/4 e^2 - 1/16 e^4 + 13/768 e^6 + 113/18432 e^8 + 2639/491520 e^10"
  G_f(2, 0,  0, e = e); print "cf. 1 - 5 e^2 + 63/8 e^4 - 155/36 e^6 + 2881/2304 e^8 - 3481/19200 e^10"
  G_f(2, 0, +1, e = e); print "cf. 0 + 49/4 e^2 - 861/16 e^4 + 21975/256 e^6 - 132635/2048 e^8 + 4654389/163840 e^10"
  G_f(2, 0, +2, e = e); print "cf. 0 + 289/4 e^4 - 1955/6 e^6 + 83551/144 e^8 - 43773/80 e^10"
  G_f(2, 0, +3, e = e); print "cf. 0 + 714025/2304 e^6 - 27483625/18432 e^8 + 587225375/196608 e^10"
  G_f(2, 0, +4, e = e); print "cf. 0 + 284089/256 e^8 - 7369791/1280 e^10"
  G_f(2, 0, +5, e = e); print "cf. 0 + 52142352409/14745600 e^10"

def test_renaud2():
  """Test orders up to cos^8 i/2 wrt. Renaud etal. (2021)."""
  i = 0.0
  print "i = ", i
  print "cf. Renaud etal. (2021)"
  F_f(2, 0, 0, i = i); print "cf. 9/4 sin^4 i/2 cos^4 i/2 = ", 9./4.*sin(i/2)**4*cos(i/2)**4
  F_f(2, 0, 1, i = i); print "cf. 1/16 (3 sin^2 i - 2)^2 = ", 1./16.*(3*sin(i)**2 - 2)**2
  F_f(2, 0, 2, i = i); print "cf. 9/4 sin^4 i/2 cos^4 i/2 = ", 9./4.*sin(i/2)**4*cos(i/2)**4
  F_f(2, 1, 0, i = i); print "cf. 9 sin^2 i/2 cos^6 i/2 = ", 9.*sin(i/2)**2*cos(i/2)**6
  F_f(2, 1, 1, i = i); print "cf. 9/16 sin^2 2 i = ", 9./16.*sin(2.*i)**2
  F_f(2, 1, 2, i = i); print "cf. 9 sin^6 i/2 cos^2 i/2 = ", 9.*sin(i/2)**6*cos(i/2)**2
  F_f(2, 2, 0, i = i); print "cf. 9 cos^8 i/2 = ", 9.*cos(i/2)**8
  F_f(2, 2, 1, i = i); print "cf. 36 sin^4 i/2 cos^4 i/2 = ", 36.*sin(i/2)**4*cos(i/2)**4
  F_f(2, 2, 2, i = i); print "cf. 9 sin^8 i/2 = ", 9.*sin(i/2)**8
  
def G_20q_sq():
  """Degree-2 secular tidal torque SQUARED, Walterova (2021), Eq. (3.1)"""

  e = 0.4
  l = 2
  m = 2
  p = 0

  f = open("G_20q_sq.dat", "w")
  f.write("# degree-2 secular tidal torque, Walterova (2021), Eq. (3.1)\n")
  f.write("# e = %f\n" % e)
  f.write("# l = %d\n" % l)
  f.write("# m = %d\n" % m)
  f.write("# p = %d\n" % p)
  f.write("#  q        (G_20q)^2    omega_lmpq\n")

  for q in xrange(-10,10+1):
    G = G_f(l, p, q, e = e, debug = False)
    f.write("%4d  %16.8f   %2d n - %2d omega\n" % (q, G**2, l-2*p+q, m))

  f.close()

def fort():
  """Output Fortran code"""
  e = 0.4
  print "e = ", e
  for q in xrange(-4,14+1):
    G_f(2, 0, q, e = e, fort = True);

def main():
#  test()
  test_renaud1()
  test_renaud2()
#  G_20q_sq()
#  fort()

if __name__ == "__main__":
  main()


